"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""

from .ibapi.object_implem import Object
from .ibapi.tag_value import TagValue
from .ibapi.order import Order


class AvailableAlgoParams(Object):

    # ! [scale_params]
    @staticmethod
    def FillScaleParams(baseOrder: Order, scaleInitLevelSize: int, scaleSubsLevelSize: int, scaleRandomPercent: bool,
                        scalePriceIncrement: float, scalePriceAdjustValue: float, scalePriceAdjustInterval: int,
                        scaleProfitOffset: float, scaleAutoReset: bool, scaleInitPosition: int, scaleInitFillQty: int):
        baseOrder.scaleInitLevelSize = scaleInitLevelSize # Initial Component Size
        baseOrder.scaleSubsLevelSize = scaleSubsLevelSize # Subsequent Comp. Size
        baseOrder.scaleRandomPercent = scaleRandomPercent # Randomize size by +/-55%
        baseOrder.scalePriceIncrement = scalePriceIncrement # Price Increment

        # Auto Price adjustment
        baseOrder.scalePriceAdjustValue = scalePriceAdjustValue # starting price by
        baseOrder.scalePriceAdjustInterval = scalePriceAdjustInterval # in seconds

        # Profit Orders
        baseOrder.scaleProfitOffset = scaleProfitOffset # Create profit taking order Profit Offset
        baseOrder.scaleAutoReset = scaleAutoReset # Restore size after taking profit
        baseOrder.scaleInitPosition = scaleInitPosition # Initial Position
        baseOrder.scaleInitFillQty = scaleInitFillQty # Filled initial Component Size
    # ! [scale_params]

    # ! [arrivalpx_params]
    @staticmethod
    def FillArrivalPriceParams(baseOrder: Order, maxPctVol: float,
                               riskAversion: str, startTime: str, endTime: str,
                               forceCompletion: bool, allowPastTime: bool):
        baseOrder.algoStrategy = "ArrivalPx"
        baseOrder.algoParams = []
        baseOrder.algoParams.append(TagValue("maxPctVol", maxPctVol))
        baseOrder.algoParams.append(TagValue("riskAversion", riskAversion))
        baseOrder.algoParams.append(TagValue("startTime", startTime))
        baseOrder.algoParams.append(TagValue("endTime", endTime))
        baseOrder.algoParams.append(TagValue("forceCompletion",
                                             int(forceCompletion)))
        baseOrder.algoParams.append(TagValue("allowPastEndTime",
                                             int(allowPastTime)))


    # ! [arrivalpx_params]


    # ! [darkice_params]
    @staticmethod
    def FillDarkIceParams(baseOrder: Order, displaySize: int, startTime: str,
                          endTime: str, allowPastEndTime: bool):
        baseOrder.algoStrategy = "DarkIce"
        baseOrder.algoParams = []
        baseOrder.algoParams.append(TagValue("displaySize", displaySize))
        baseOrder.algoParams.append(TagValue("startTime", startTime))
        baseOrder.algoParams.append(TagValue("endTime", endTime))
        baseOrder.algoParams.append(TagValue("allowPastEndTime",
                                             int(allowPastEndTime)))


    # ! [darkice_params]


    # ! [pctvol_params]
    @staticmethod
    def FillPctVolParams(baseOrder: Order, pctVol: float, startTime: str,
                         endTime: str, noTakeLiq: bool):
        baseOrder.algoStrategy = "PctVol"
        baseOrder.algoParams = []
        baseOrder.algoParams.append(TagValue("pctVol", pctVol))
        baseOrder.algoParams.append(TagValue("startTime", startTime))
        baseOrder.algoParams.append(TagValue("endTime", endTime))
        baseOrder.algoParams.append(TagValue("noTakeLiq", int(noTakeLiq)))


    # ! [pctvol_params]


    # ! [twap_params]
    @staticmethod
    def FillTwapParams(baseOrder: Order, StrategyName: str, startTime: str,
                       endTime: str, allowPastEndTime: bool):
        baseOrder.algoStrategy = "Twap"
        baseOrder.algoParams = []
        baseOrder.algoParams.append(TagValue("StrategyName", StrategyName))
        baseOrder.algoParams.append(TagValue("startTime", startTime))
        baseOrder.algoParams.append(TagValue("endTime", endTime))
        baseOrder.algoParams.append(TagValue("allowPastEndTime",
                                             int(allowPastEndTime)))

    # ! [twap_params]


    # ! [vwap_params]
    @staticmethod
    def FillVwapParams(baseOrder: Order, maxPctVol: float, startTime: str,
                       endTime: str, allowPastEndTime: bool, noTakeLiq: bool):
        baseOrder.algoStrategy = "Vwap"
        baseOrder.algoParams = []
        baseOrder.algoParams.append(TagValue("maxPctVol", maxPctVol))
        baseOrder.algoParams.append(TagValue("startTime", startTime))
        baseOrder.algoParams.append(TagValue("endTime", endTime))
        baseOrder.algoParams.append(TagValue("allowPastEndTime",
                                             int(allowPastEndTime)))
        baseOrder.algoParams.append(TagValue("noTakeLiq", int(noTakeLiq)))


    # ! [vwap_params]


    # ! [ad_params]
    @staticmethod
    def FillAccumulateDistributeParams(baseOrder: Order, componentSize: int,
                                       timeBetweenOrders: int, randomizeTime20: bool, randomizeSize55: bool,
                                       giveUp: int, catchUp: bool, waitForFill: bool, startTime: str,
                                       endTime: str):
        baseOrder.algoStrategy = "AD"
        baseOrder.algoParams = []
        baseOrder.algoParams.append(TagValue("componentSize", componentSize))
        baseOrder.algoParams.append(TagValue("timeBetweenOrders", timeBetweenOrders))
        baseOrder.algoParams.append(TagValue("randomizeTime20",
                                             int(randomizeTime20)))
        baseOrder.algoParams.append(TagValue("randomizeSize55",
                                             int(randomizeSize55)))
        baseOrder.algoParams.append(TagValue("giveUp", giveUp))
        baseOrder.algoParams.append(TagValue("catchUp", int(catchUp)))
        baseOrder.algoParams.append(TagValue("waitForFill", int(waitForFill)))
        baseOrder.algoParams.append(TagValue("activeTimeStart", startTime))
        baseOrder.algoParams.append(TagValue("activeTimeEnd", endTime))

    # ! [ad_params]


    # ! [balanceimpactrisk_params]
    @staticmethod
    def FillBalanceImpactRiskParams(baseOrder: Order, maxPctVol: float,
                                    riskAversion: str, forceCompletion: bool):
        baseOrder.algoStrategy = "BalanceImpactRisk"
        baseOrder.algoParams = []
        baseOrder.algoParams.append(TagValue("maxPctVol", maxPctVol))
        baseOrder.algoParams.append(TagValue("riskAversion", riskAversion))
        baseOrder.algoParams.append(TagValue("forceCompletion",
                                             int(forceCompletion)))

    # ! [balanceimpactrisk_params]


    # ! [minimpact_params]
    @staticmethod
    def FillMinImpactParams(baseOrder: Order, maxPctVol: float):
        baseOrder.algoStrategy = "MinImpact"
        baseOrder.algoParams = []
        baseOrder.algoParams.append(TagValue("maxPctVol", maxPctVol))

    # ! [minimpact_params]


    # ! [adaptive_params]
    @staticmethod
    def FillAdaptiveParams(baseOrder: Order, priority: str):
        baseOrder.algoStrategy = "Adaptive"
        baseOrder.algoParams = []
        baseOrder.algoParams.append(TagValue("adaptivePriority", priority))

    # ! [adaptive_params]

    # ! [closepx_params]
    @staticmethod
    def FillClosePriceParams(baseOrder: Order, maxPctVol: float, riskAversion: str,
                             startTime: str, forceCompletion: bool):
        baseOrder.algoStrategy = "ClosePx"
        baseOrder.algoParams = []
        baseOrder.algoParams.append(TagValue("maxPctVol", maxPctVol))
        baseOrder.algoParams.append(TagValue("riskAversion", riskAversion))
        baseOrder.algoParams.append(TagValue("startTime", startTime))
        baseOrder.algoParams.append(TagValue("forceCompletion", int(forceCompletion)))


    # ! [closepx_params]


    # ! [pctvolpx_params]
    @staticmethod
    def FillPriceVariantPctVolParams(baseOrder: Order, pctVol: float,
                                     deltaPctVol: float, minPctVol4Px: float,
                                     maxPctVol4Px: float, startTime: str,
                                     endTime: str, noTakeLiq: bool):
        baseOrder.algoStrategy = "PctVolPx"
        baseOrder.algoParams = []
        baseOrder.algoParams.append(TagValue("pctVol", pctVol))
        baseOrder.algoParams.append(TagValue("deltaPctVol", deltaPctVol))
        baseOrder.algoParams.append(TagValue("minPctVol4Px", minPctVol4Px))
        baseOrder.algoParams.append(TagValue("maxPctVol4Px", maxPctVol4Px))
        baseOrder.algoParams.append(TagValue("startTime", startTime))
        baseOrder.algoParams.append(TagValue("endTime", endTime))
        baseOrder.algoParams.append(TagValue("noTakeLiq", int(noTakeLiq)))


    # ! [pctvolpx_params]


    # ! [pctvolsz_params]
    @staticmethod
    def FillSizeVariantPctVolParams(baseOrder: Order, startPctVol: float,
                                    endPctVol: float, startTime: str,
                                    endTime: str, noTakeLiq: bool):
        baseOrder.algoStrategy = "PctVolSz"
        baseOrder.algoParams = []
        baseOrder.algoParams.append(TagValue("startPctVol", startPctVol))
        baseOrder.algoParams.append(TagValue("endPctVol", endPctVol))
        baseOrder.algoParams.append(TagValue("startTime", startTime))
        baseOrder.algoParams.append(TagValue("endTime", endTime))
        baseOrder.algoParams.append(TagValue("noTakeLiq", int(noTakeLiq)))


    # ! [pctvolsz_params]


    # ! [pctvoltm_params]
    @staticmethod
    def FillTimeVariantPctVolParams(baseOrder: Order, startPctVol: float,
                                    endPctVol: float, startTime: str,
                                    endTime: str, noTakeLiq: bool):
        baseOrder.algoStrategy = "PctVolTm"
        baseOrder.algoParams = []
        baseOrder.algoParams.append(TagValue("startPctVol", startPctVol))
        baseOrder.algoParams.append(TagValue("endPctVol", endPctVol))
        baseOrder.algoParams.append(TagValue("startTime", startTime))
        baseOrder.algoParams.append(TagValue("endTime", endTime))
        baseOrder.algoParams.append(TagValue("noTakeLiq", int(noTakeLiq)))

    # ! [pctvoltm_params]

    # ! [jefferies_vwap_params]
    @staticmethod
    def FillJefferiesVWAPParams(baseOrder: Order, startTime: str,
                                endTime: str, relativeLimit: float,
                                maxVolumeRate: float, excludeAuctions: str,
                                triggerPrice: float, wowPrice: float,
                                minFillSize: int, wowOrderPct: float,
                                wowMode: str, isBuyBack: bool, wowReference: str):
        # must be direct-routed to "JEFFALGO"
        baseOrder.algoStrategy = "VWAP"
        baseOrder.algoParams = []
        baseOrder.algoParams.append(TagValue("StartTime", startTime))
        baseOrder.algoParams.append(TagValue("EndTime", endTime))
        baseOrder.algoParams.append(TagValue("RelativeLimit", relativeLimit))
        baseOrder.algoParams.append(TagValue("MaxVolumeRate", maxVolumeRate))
        baseOrder.algoParams.append(TagValue("ExcludeAuctions", excludeAuctions))
        baseOrder.algoParams.append(TagValue("TriggerPrice", triggerPrice))
        baseOrder.algoParams.append(TagValue("WowPrice", wowPrice))
        baseOrder.algoParams.append(TagValue("MinFillSize", minFillSize))
        baseOrder.algoParams.append(TagValue("WowOrderPct", wowOrderPct))
        baseOrder.algoParams.append(TagValue("WowMode", wowMode))
        baseOrder.algoParams.append(TagValue("IsBuyBack", int(isBuyBack)))
        baseOrder.algoParams.append(TagValue("WowReference", wowReference))
    # ! [jefferies_vwap_params]

    # ! [csfb_inline_params]
    @staticmethod
    def FillCSFBInlineParams(baseOrder: Order, startTime: str,
                             endTime: str, execStyle: str,
                             minPercent: int, maxPercent: int,
                             displaySize: int, auction: str,
                             blockFinder: bool, blockPrice: float,
                             minBlockSize: int, maxBlockSize: int, iWouldPrice: float):
        # must be direct-routed to "CSFBALGO"
        baseOrder.algoStrategy = "INLINE"
        baseOrder.algoParams = []
        baseOrder.algoParams.append(TagValue("StartTime", startTime))
        baseOrder.algoParams.append(TagValue("EndTime", endTime))
        baseOrder.algoParams.append(TagValue("ExecStyle", execStyle))
        baseOrder.algoParams.append(TagValue("MinPercent", minPercent))
        baseOrder.algoParams.append(TagValue("MaxPercent", maxPercent))
        baseOrder.algoParams.append(TagValue("DisplaySize", displaySize))
        baseOrder.algoParams.append(TagValue("Auction", auction))
        baseOrder.algoParams.append(TagValue("BlockFinder", int(blockFinder)))
        baseOrder.algoParams.append(TagValue("BlockPrice", blockPrice))
        baseOrder.algoParams.append(TagValue("MinBlockSize", minBlockSize))
        baseOrder.algoParams.append(TagValue("MaxBlockSize", maxBlockSize))
        baseOrder.algoParams.append(TagValue("IWouldPrice", iWouldPrice))
    # ! [csfb_inline_params]

    # ! [qbalgo_strobe_params]
    @staticmethod
    def FillQBAlgoInLineParams(baseOrder: Order, startTime: str, 
                               endTime: str, duration: float, 
                               benchmark: str, percentVolume: float, 
                               noCleanUp: bool):
        # must be direct-routed to "QBALGO"
        baseOrder.algoStrategy = "STROBE"
        baseOrder.algoParams = []
        baseOrder.algoParams.append(TagValue("StartTime", startTime))
        baseOrder.algoParams.append(TagValue("EndTime", endTime))
        #This example uses endTime instead of duration
        #baseOrder.algoParams.append(TagValue("Duration", str(duration)))
        baseOrder.algoParams.append(TagValue("Benchmark", benchmark)) 
        baseOrder.algoParams.append(TagValue("PercentVolume", str(percentVolume)))
        baseOrder.algoParams.append(TagValue("NoCleanUp", int(noCleanUp)))
    # ! [qbalgo_strobe_params]


def Test():
    av = AvailableAlgoParams() # @UnusedVariable


if "__main__" == __name__:
    Test()



"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""


from .object_implem import Object
from .common import UNSET_INTEGER, UNSET_DOUBLE


class ScanData(Object):
    def __init__(self, contract = None, rank = 0, distance = "", benchmark = "", projection = "", legsStr = ""):
        self.contract = contract
        self.rank = rank
        self.distance = distance
        self.benchmark = benchmark
        self.projection = projection
        self.legsStr = legsStr

    def __str__(self):
        return "Rank: %d, Symbol: %s, SecType: %s, Currency: %s, Distance: %s, Benchmark: %s, Projection: %s, Legs String: %s" % (self.rank, 
            self.contract.symbol, self.contract.secType, self.contract.currency, self.distance,
            self.benchmark, self.projection, self.legsStr)


NO_ROW_NUMBER_SPECIFIED = -1


class ScannerSubscription(Object):
    def __init__(self):
        self.numberOfRows = NO_ROW_NUMBER_SPECIFIED
        self.instrument = ""
        self.locationCode = ""
        self.scanCode =  ""
        self.abovePrice = UNSET_DOUBLE
        self.belowPrice = UNSET_DOUBLE
        self.aboveVolume = UNSET_INTEGER
        self.marketCapAbove = UNSET_DOUBLE
        self.marketCapBelow = UNSET_DOUBLE
        self.moodyRatingAbove =  ""
        self.moodyRatingBelow =  ""
        self.spRatingAbove =  ""
        self.spRatingBelow =  ""
        self.maturityDateAbove =  ""
        self.maturityDateBelow =  ""
        self.couponRateAbove = UNSET_DOUBLE
        self.couponRateBelow = UNSET_DOUBLE 
        self.excludeConvertible = False
        self.averageOptionVolumeAbove = UNSET_INTEGER
        self.scannerSettingPairs =  ""
        self.stockTypeFilter =  ""

    def __str__(self):
        s = "Instrument: %s, LocationCode: %s, ScanCode: %s" % (self.instrument, self.locationCode, self.scanCode)
        return s




"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""


class AccountSummaryTags:
    AccountType = "AccountType"
    NetLiquidation = "NetLiquidation"
    TotalCashValue = "TotalCashValue"
    SettledCash = "SettledCash"
    AccruedCash = "AccruedCash"
    BuyingPower = "BuyingPower"
    EquityWithLoanValue = "EquityWithLoanValue"
    PreviousDayEquityWithLoanValue = "PreviousDayEquityWithLoanValue"
    GrossPositionValue = "GrossPositionValue"
    ReqTEquity = "ReqTEquity"
    ReqTMargin = "ReqTMargin"
    SMA = "SMA"
    InitMarginReq = "InitMarginReq"
    MaintMarginReq = "MaintMarginReq"
    AvailableFunds = "AvailableFunds"
    ExcessLiquidity = "ExcessLiquidity"
    Cushion = "Cushion"
    FullInitMarginReq = "FullInitMarginReq"
    FullMaintMarginReq = "FullMaintMarginReq"
    FullAvailableFunds = "FullAvailableFunds"
    FullExcessLiquidity = "FullExcessLiquidity"
    LookAheadNextChange = "LookAheadNextChange"
    LookAheadInitMarginReq = "LookAheadInitMarginReq"
    LookAheadMaintMarginReq = "LookAheadMaintMarginReq"
    LookAheadAvailableFunds = "LookAheadAvailableFunds"
    LookAheadExcessLiquidity = "LookAheadExcessLiquidity"
    HighestSeverity = "HighestSeverity"
    DayTradesRemaining = "DayTradesRemaining"
    Leverage = "Leverage"

    AllTags = ",".join((AccountType, NetLiquidation, TotalCashValue,
        SettledCash, AccruedCash, BuyingPower, EquityWithLoanValue,
        PreviousDayEquityWithLoanValue, GrossPositionValue, ReqTEquity,
        ReqTMargin, SMA, InitMarginReq, MaintMarginReq, AvailableFunds, 
        ExcessLiquidity , Cushion, FullInitMarginReq, FullMaintMarginReq, 
        FullAvailableFunds, FullExcessLiquidity,
        LookAheadNextChange, LookAheadInitMarginReq, LookAheadMaintMarginReq,
        LookAheadAvailableFunds, LookAheadExcessLiquidity, HighestSeverity,
        DayTradesRemaining, Leverage))





# todo 1. 对于查询的持仓，空的也要推送空的，否则orderplit无法回调.  这对于http请求很容易实现，但是如果是websocket回调，也许空的不会回调？例如ibk
# todo 2. 持仓查询已经完成了direction过滤，order查询还没有经过参数过滤

import pandas as pd
from datetime import datetime
from typing import Union, Tuple, Optional
import itertools
from ks_trade_api.object import ContractData, MyAccountData, ErrorData, MyPositionData, MyTradeData, MyOrderData, MyOrderRequest, MySubscribeRequest
from ks_trade_api.constant import (
    Product as ksProduct,
    Currency as KsCurrency,
    Exchange as Exchange,
    Direction as KsDirection, 
    OrderType as ksOrderType, 
    Direction as KsDirection,
    SubscribeType as KsSubscribeType,
    Offset as KsOffset, 
    TimeInForce as KsTimeInForce,
    TradingHours as KsTradingHours,
    ErrorCode as KsErrorCode,
    Status as KsStatus,
    RetCode as KsRetCode,
    RET_OK as KS_RET_OK, 
    RET_ASYNC as KS_RET_ASYNC,
    RET_ERROR as KS_RET_ERROR, 
    CHINA_TZ,
    US_EASTERN_TZ
)
from .ib_api import IbApi
from dateutil.parser import parse
from ks_trade_api.base_trade_api import BaseTradeApi
from ks_trade_api.utility import extract_vt_symbol, extract_vt_orderid
import sys
from decimal import Decimal
import uuid
from logging import DEBUG, WARNING, ERROR
from .ib_api import CURRENCY_VT2IB, raw_symbol_ks2my, raw_symbol_my2ks

RATES_INTERVAL: int = 30

class OrderType():
    LMT = 'LMT'
    MKT = 'MKT'

class TrdSide():
    BUY = 'BUY'
    SELL = 'SELL'


SUBTYPE_KS2MY = {
    KsSubscribeType.USER_ORDER: KsSubscribeType.USER_ORDER,
    KsSubscribeType.USER_TRADE: KsSubscribeType.USER_TRADE,
    KsSubscribeType.USER_POSITION: KsSubscribeType.USER_POSITION,
}

MARKET_KS2MY = {
    Exchange.SEHK: 'HK',
    Exchange.SMART: 'US'
}

MARKET_MY2KS = { v:k for k,v in MARKET_KS2MY.items() }

def symbol_ks2my(vt_symbol: str):
    if not vt_symbol:
        return ''
    symbol, ks_exchange = extract_vt_symbol(vt_symbol)
    return f'{MARKET_KS2MY.get(ks_exchange)}.{symbol}'

ORDERTYPE_MY2KS = {
    OrderType.LMT: ksOrderType.LIMIT,
    OrderType.MKT: ksOrderType.MARKET
}

ORDERTYPE_KS2MY = {v:k for k,v in ORDERTYPE_MY2KS.items()}


SIDE_KS2MY = {
    f'{KsDirection.LONG.value},{KsOffset.OPEN.value}': TrdSide.BUY,
    f'{KsDirection.SHORT.value},{KsOffset.CLOSE.value}': TrdSide.SELL,
    f'{KsDirection.SHORT.value},{KsOffset.CLOSETODAY.value}': TrdSide.SELL,
    f'{KsDirection.SHORT.value},{KsOffset.CLOSEYESTERDAY.value}': TrdSide.SELL,

    f'{KsDirection.SHORT.value},{KsOffset.OPEN.value}': TrdSide.SELL,
    f'{KsDirection.LONG.value},{KsOffset.CLOSE.value}': TrdSide.BUY,
    f'{KsDirection.LONG.value},{KsOffset.CLOSETODAY.value}': TrdSide.BUY,
    f'{KsDirection.LONG.value},{KsOffset.CLOSEYESTERDAY.value}': TrdSide.BUY,
}

def side_ks2my(direction: KsDirection, offset: KsOffset):
    key = f'{direction.value},{offset.value}'
    return SIDE_KS2MY.get(key)

def sides_ks2my(directions: list[KsDirection], offsets: list[KsOffset]):
    sides_map = {}
    combinations = itertools.product(directions, offsets)
    for direction, offset in combinations:
        sides_map[side_ks2my(direction, offset)] = 1
    sides = list(sides_map.keys())
    return sides

def side_my2ks(side: TrdSide):
    # 为啥原来会是对的？
    if side == TrdSide.BUY:
        direction = KsDirection.LONG
        offset = KsOffset.OPEN
    else:
        direction = KsDirection.SHORT
        offset = KsOffset.CLOSE
    return direction, offset



# 定义一个自定义错误类
class MyError(Exception):
    def __init__(self, message):
        super().__init__(message)
        self.message = message
        # futu没有错误代码。。只能用文字先替代
        if '购买力不足' in message:
            self.code = KsErrorCode.BUY_POWER_EXCEEDED
        elif '频率太高' in message:
            self.code = KsErrorCode.RATE_LIMITS_EXCEEDED

class KsIbkrApi(BaseTradeApi):
    gateway_name: str = "KS_IBKR"

    ERROR_CODE_MY2KS: dict = {
        KsErrorCode.BUY_POWER_EXCEEDED: KsErrorCode.BUY_POWER_EXCEEDED,
        KsErrorCode.RATE_LIMITS_EXCEEDED: KsErrorCode.RATE_LIMITS_EXCEEDED
    }

    def __init__(self, setting: dict):
        self.port = setting.get('port', 7497)
        self.client_id = setting.get('client_id', 666)
        dd_secret = setting.get('dd_secret')
        dd_token = setting.get('dd_token')
        gatway_name: str = setting.get('gateway_name', self.gateway_name)

        self.vt_orderid_acc_id_map: dict = {} # 记录订单的account，因为订单撤销需要账号
        
        super().__init__(gateway_name=gatway_name, dd_secret=dd_secret, dd_token=dd_token, setting=setting)

        self.positions = []
        self.query_position_params = {}
        
        self.init_handlers()



    # 初始化行回调和订单回调
    def init_handlers(self):
        self.ctx = IbApi(self, self.port, client_id = self.client_id)
    
    def on_contract(self, contract):
        pass   

    def on_positions(self, positions: list[MyPositionData]) -> None:
        pass
        

    # 订阅行情
    def subscribe(self, vt_symbols, vt_subtype_list, product: ksProduct = ksProduct.EQUITY, extended_time=True) -> tuple[KsRetCode, Optional[ErrorData]]:
        for vt_symbol in vt_symbols:
            symbol, exchange = extract_vt_symbol(vt_symbol)
            req = MySubscribeRequest(symbol=symbol, exchange=exchange, types=vt_subtype_list, product=product)
            self.ctx.subscribe(req)

    def query_contract(self, vt_symbol: str) -> tuple[KsRetCode, ContractData]:
        try:
            self.ctx.query_contract(vt_symbol)
        except Exception as e:
            error = self.get_error(vt_symbol, e=e)
            self.send_dd(error.msg, f'合约查询错误')
            return KS_RET_ERROR, error
        
        return KS_RET_ASYNC, None
 
    
    # 下单
    # @RateLimitChecker(RATES_INTERVAL)
    def send_order(
            self, 
            vt_symbol: str,
            price: Decimal,
            volume: Decimal,
            type: ksOrderType = ksOrderType.LIMIT,
            direction: KsDirection = KsDirection.LONG,
            offset: KsOffset = KsOffset.OPEN,
            time_in_force: KsTimeInForce = KsTimeInForce.GTC,
            trading_hours: KsTradingHours = KsTradingHours.RTH,
            reference: str = '',
            product: ksProduct = ksProduct.EQUITY
    ) -> Tuple[KsRetCode, Union[str, ErrorData]]:
        symbol, exchange = extract_vt_symbol(vt_symbol)
        order_request: MyOrderRequest = MyOrderRequest(
            symbol=symbol,
            exchange=exchange,
            direction=direction,
            offset=offset,
            type=type,
            price=price,
            volume=volume,
            time_in_force=time_in_force,
            reference=reference,
            product=product
        )
        try:
            vt_orderid =self.ctx.send_order(order_request)

            self.log({
                'vt_symbol': vt_symbol,
                'direction': direction,
                'offset': offset,
                
                'symbol': symbol,
                'price': price,
                'volume': volume,
                'type': type,
                'direction': direction,
                'offset': offset,
                'time_in_force': time_in_force,
                'trading_hours': trading_hours,
                'reference': reference,
                'vt_orderid': vt_orderid
            })
            return KS_RET_OK, vt_orderid
        except Exception as e:
            error = self.get_error(params={
                'symbol': symbol,
                'price': price,
                'volume': volume,
                'type': type,
                'direction': direction,
                'offset': offset,
                'time_in_force': time_in_force,
                'trading_hours': trading_hours,
                'reference': reference
            }, e=e)
            return KS_RET_ERROR, error
        
    # My.add 直接向服务器请求合约数据
    # @RateLimitChecker(RATES_INTERVAL)
    def request_cancel_orders(
            self,
            vt_symbol: Optional[str] = None,
            direction: KsDirection = None,
            offset: KsOffset = None
        ) -> tuple[KsRetCode,  list[MyOrderData]]:
        ret = KS_RET_OK
        orders = []
        
        ret_query, open_orders = self.query_open_orders(vt_symbol=vt_symbol, direction=direction, offset=offset)
        if ret_query == KS_RET_ASYNC:
            # todo 异步尚未处理好
            return ret_query, open_orders
        
        if ret_query == KS_RET_OK:  
            for order in open_orders:
                order: MyOrderData
                ret_cancel, cancel_res = self.cancel_order(order.vt_orderid)
                # todo这里没有处理异步
                if ret_cancel == KS_RET_OK:
                    order.status = KsStatus.CANCELLED
                    orders.append(order)
                else:
                    ret = ret_cancel
                    orders = cancel_res
                    break
        else:
            ret = ret_query
            orders = open_orders
            

        return ret, orders

    # 撤单
    # @RateLimitChecker(RATES_INTERVAL)
    def cancel_order(self, vt_orderid: str) -> Tuple[KsRetCode, Optional[ErrorData]]:
        self.log({'vt_orderid': vt_orderid}, level=DEBUG)
        gateway_name, orderid = extract_vt_orderid(vt_orderid)
        try:
            self.ctx.cancel_order(orderid)
        except Exception as e:
            error = self.get_error(orderid, e=e)
            return KS_RET_ERROR, error

        return KS_RET_OK, vt_orderid

    # 获取账号信息
    def query_account(self, currencies: list[KsCurrency] = []) -> tuple[KsRetCode, Union[MyAccountData, ErrorData]]:
        # IB初始化会更新account，不需要主动查询 有更新自动推送，没有更新三分钟推送
        # self.ctx.reqAccountUpdates()
        return KS_RET_ASYNC, None
        accounts: dict[MyAccountData] = []

        try:
            data = self.ctx.accountSummary()
        except Exception as e:
            error = self.get_error(currency=currency, e=e)
            return KS_RET_ERROR, error
        
        # if len(data):
        #     balance = next((float(x.value) for x in data if x.tag == 'EquityWithLoanValue'), None)
        #     available = next((float(x.value) for x in data if x.tag == 'BuyingPower'), None)
        #     account: MyAccountData = MyAccountData(
        #         accountid='',
        #         balance=balance,
        #         frozen=0,
        #         currency=KsCurrency.HKD,
        #         gateway_name=self.gateway_name,
        #     )
        #     account.available = available
        #     accounts.append(account)

        return KS_RET_ASYNC, None
        


    # 获取持仓信息
    # @RateLimitChecker(RATES_INTERVAL)
    def query_position(self, vt_symbols=[], directions: list[KsDirection] = []):
        self.positions = []
        self.query_position_params = {'vt_symbols': vt_symbols, 'directions': directions}
        try:
            self.ctx.client.reqPositions()
        except Exception as e:
            error = self.get_error(vt_symbols, e=e)
            self.send_dd(error.msg, f'持仓查询错误')
            return KS_RET_ERROR, error
        
        return KS_RET_ASYNC, None
        
        positions = []
        if len(data):
            for position_data in data:
                # todo 这里持仓返回的exchange并不是smart，而是具体的，这导致跟我们配置的smart不符合，所以要重设为smart
                contract = position_data.contract
                contract.exchange = ''
                self.ctx.qualifyContracts(contract)
                
                if contract.secType == 'CASH':
                    symbol = contract.pair()
                else:
                    symbol = contract.symbol

                exchange = Exchange(contract.exchange)
                direction = KsDirection.NET
                position = MyPositionData(
                    symbol=symbol,
                    exchange=exchange,
                    direction=direction,
                    price=Decimal(str(position_data.avgCost)),
                    volume=Decimal(str(position_data.position)),
                    available=Decimal(str(position_data.position)),
                    gateway_name=self.gateway_name
                )
                positions.append(position)

        # 去除多查询的持仓
        if vt_symbols:
            positions = [x for x in positions if x.vt_symbol in vt_symbols]

        # 如果当天清仓再开仓，会有两条持仓记录，我们直接把0的记录删除
        positions = [x for x in positions if x.volume]
        

        # 补齐空持仓
        ret_ks_symbols = [x.vt_symbol for x in positions]
        lack_ks_symbols = [x for x in vt_symbols if not x in ret_ks_symbols]
        for lack_ks_symbol in lack_ks_symbols:
            if not lack_ks_symbol:
                continue
            symbol, exchange = extract_vt_symbol(lack_ks_symbol)
            lack_postion = MyPositionData(symbol=symbol, exchange=exchange, direction=KsDirection.NET, gateway_name=self.gateway_name)
            positions.append(lack_postion)

        return KS_RET_OK, positions
    
    def on_position_(self, position: MyPositionData) -> None:
        self.positions.append(position)

    def on_positions_end_(self) -> None:
        # 去除多查询的持仓
        vt_symbols: list[str] = self.query_position_params.get('vt_symbols', [])
        directions: list[KsDirection] = self.query_position_params.get('directions', [])
        if vt_symbols:
            self.positions = [x for x in self.positions if x.vt_symbol in vt_symbols]
        # 补齐空持仓
        ret_ks_symbols = [x.vt_symbol for x in self.positions]
        lack_ks_symbols = [x for x in vt_symbols if not x in ret_ks_symbols]
        for lack_ks_symbol in lack_ks_symbols:
            if not lack_ks_symbol:
                continue
            symbol, exchange = extract_vt_symbol(lack_ks_symbol)
            lack_postion = MyPositionData(symbol=symbol, exchange=exchange, direction=KsDirection.NET, gateway_name=self.gateway_name)
            self.positions.append(lack_postion)
        # 过滤direction
        if directions:
            self.positions = [x for x in self.positions if x.direction in directions]

        for position in self.positions:
            self.on_position(position)

        if self.setting.get('log.position', False):
            self.log(self.positions, tag="持仓回调", level=DEBUG)
    
        self.on_positions_end()

    # 获取今日订单
    def query_orders(self, 
        vt_symbol: Optional[str] = None, 
        direction: Optional[KsDirection] = None, 
        offset: Optional[KsOffset] = None,
        status: Optional[list[KsStatus]] = None,
        orderid: Optional[str] = None,
        reference: Optional[str] = None 
    ) -> tuple[KsRetCode, Union[list[MyOrderData], ErrorData]]:
        try:
            # 获取已完成和未完成订单
            self.ctx.client.reqOpenOrders()
            self.ctx.client.reqCompletedOrders(apiOnly=True) 
        except Exception as e:   
            error = self.get_error(vt_symbol, direction, offset, status, orderid, reference, e=e)
            return KS_RET_ERROR, error
        
        # orders = [self.order_data_my2ks(x) for x in orders]
        # # todo 过滤status
        # # 如果指定了代码则过滤代码
        # if vt_symbol:
        #     orders = [x for x in orders if x.vt_symbol == vt_symbol]
        # if direction:
        #     orders = [x for x in orders if x.direction == direction]
        # if direction:
        #     orders = [x for x in orders if x.offset == offset]
        # if direction:
        #     orders = [x for x in orders if x.status in status]
        # if orderid:
        #     orders = [x for x in orders if x.orderid == orderid]
        # if reference:
        #     orders = [x for x in orders if x.reference == reference]

        return KS_RET_ASYNC, None

        
    # 获取今日订单 # todo get_orders没有实现
    def query_open_orders(self, 
            vt_symbol: Optional[str]=None, 
            direction: Optional[KsDirection] = None, 
            offset: Optional[KsOffset] = None,
            status: Optional[list[KsStatus]] = None,
            orderid: Optional[str] = None,
            reference: Optional[str] = None
    ) -> tuple[KsRetCode, Union[list[MyOrderData], ErrorData]]:
        try:
            # 获取未完成订单
            self.ctx.client.reqOpenOrders()
        except Exception as e:   
            error = self.get_error(vt_symbol, direction, offset, status, orderid, reference, e=e)
            return KS_RET_ERROR, error
        
        # orders = [self.order_data_my2ks(x) for x in orders]
        # # todo 过滤status
        # # 如果指定了代码则过滤代码
        # if vt_symbol:
        #     orders = [x for x in orders if x.vt_symbol == vt_symbol]
        # if direction:
        #     orders = [x for x in orders if x.direction == direction]
        # if direction:
        #     orders = [x for x in orders if x.offset == offset]
        # if direction:
        #     orders = [x for x in orders if x.status in status]
        # if orderid:
        #     orders = [x for x in orders if x.orderid == orderid]
        # if reference:
        #     orders = [x for x in orders if x.reference == reference]

        return KS_RET_ASYNC, None


    def close(self):
        self.log('ibkr connection closed!')
        self.ctx.close()


        


"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""

import sys
# todo!!! 移除这一行不知道有没有影响交易，ibkr交易还没回测
# import ibapi
import math

from decimal import Decimal
from .enum_implem import Enum
from .object_implem import Object


NO_VALID_ID = -1
MAX_MSG_LEN = 0xFFFFFF # 16Mb - 1byte

UNSET_INTEGER = 2 ** 31 - 1
UNSET_DOUBLE = sys.float_info.max
UNSET_LONG = 2 ** 63 - 1
UNSET_DECIMAL = Decimal(2 ** 127 - 1)
DOUBLE_INFINITY = math.inf

INFINITY_STR = "Infinity"

TickerId = int
OrderId  = int
TagValueList = list

FaDataType = int
FaDataTypeEnum = Enum("N/A", "GROUPS", "PROFILES", "ALIASES")

MarketDataType = int
MarketDataTypeEnum = Enum("N/A", "REALTIME", "FROZEN", "DELAYED", "DELAYED_FROZEN")

Liquidities = int
LiquiditiesEnum = Enum("None", "Added", "Remove", "RoudedOut")

SetOfString = set
SetOfFloat = set
ListOfOrder = list
ListOfFamilyCode = list
ListOfContractDescription = list
ListOfDepthExchanges = list
ListOfNewsProviders = list
SmartComponentMap = dict
HistogramDataList = list
ListOfPriceIncrements = list
ListOfHistoricalTick = list
ListOfHistoricalTickBidAsk = list
ListOfHistoricalTickLast = list
ListOfHistoricalSessions = list


class BarData(Object):
    def __init__(self):
        self.date = ""
        self.open = 0.
        self.high = 0.
        self.low = 0.
        self.close = 0.
        self.volume = UNSET_DECIMAL
        self.wap = UNSET_DECIMAL
        self.barCount = 0

    def __str__(self):
        return "Date: %s, Open: %s, High: %s, Low: %s, Close: %s, Volume: %s, WAP: %s, BarCount: %s" % (self.date, ibapi.utils.floatMaxString(self.open), 
            ibapi.utils.floatMaxString(self.high), ibapi.utils.floatMaxString(self.low), ibapi.utils.floatMaxString(self.close), 
            ibapi.utils.decimalMaxString(self.volume), ibapi.utils.decimalMaxString(self.wap), ibapi.utils.intMaxString(self.barCount))


class RealTimeBar(Object):
    def __init__(self, time = 0, endTime = -1, open_ = 0., high = 0., low = 0., close = 0., volume = UNSET_DECIMAL, wap = UNSET_DECIMAL, count = 0):
        self.time = time
        self.endTime = endTime
        self.open_ = open_
        self.high = high
        self.low = low
        self.close = close
        self.volume = volume
        self.wap = wap
        self.count = count

    def __str__(self):
        return "Time: %s, Open: %s, High: %s, Low: %s, Close: %s, Volume: %s, WAP: %s, Count: %s" % (ibapi.utils.intMaxString(self.time), 
            ibapi.utils.floatMaxString(self.open), ibapi.utils.floatMaxString(self.high), ibapi.utils.floatMaxString(self.low), 
            ibapi.utils.floatMaxString(self.close), ibapi.utils.decimalMaxString(self.volume), ibapi.utils.decimalMaxString(self.wap), 
            ibapi.utils.intMaxString(self.count))


class HistogramData(Object):
    def __init__(self):
        self.price = 0.
        self.size = UNSET_DECIMAL

    def __str__(self):
        return "Price: %s, Size: %s" % (ibapi.utils.floatMaxString(self.price), ibapi.utils.decimalMaxString(self.size))


class NewsProvider(Object):
    def __init__(self):
        self.code = ""
        self.name = ""

    def __str__(self):
        return "Code: %s, Name: %s" % (self.code, self.name)


class DepthMktDataDescription(Object):
    def __init__(self):
        self.exchange = ""
        self.secType = ""
        self.listingExch = ""
        self.serviceDataType = ""
        self.aggGroup = UNSET_INTEGER

    def __str__(self):
        if (self.aggGroup!= UNSET_INTEGER):
            aggGroup = self.aggGroup
        else:
            aggGroup = ""
        return "Exchange: %s, SecType: %s, ListingExchange: %s, ServiceDataType: %s, AggGroup: %s, " % (self.exchange, self.secType, 
            self.listingExch,self.serviceDataType, ibapi.utils.intMaxString(aggGroup))


class SmartComponent(Object):
    def __init__(self):
        self.bitNumber = 0
        self.exchange = ""
        self.exchangeLetter = ""

    def __str__(self):
        return "BitNumber: %d, Exchange: %s, ExchangeLetter: %s" % (self.bitNumber, self.exchange, self.exchangeLetter)


class TickAttrib(Object):
    def __init__(self):
        self.canAutoExecute = False
        self.pastLimit = False
        self.preOpen = False

    def __str__(self):
        return "CanAutoExecute: %d, PastLimit: %d, PreOpen: %d" % (self.canAutoExecute, self.pastLimit, self.preOpen)


class TickAttribBidAsk(Object):
    def __init__(self):
        self.bidPastLow = False
        self.askPastHigh = False

    def __str__(self):
        return "BidPastLow: %d, AskPastHigh: %d" % (self.bidPastLow, self.askPastHigh)


class TickAttribLast(Object):
    def __init__(self):
        self.pastLimit = False
        self.unreported = False

    def __str__(self):
        return "PastLimit: %d, Unreported: %d" % (self.pastLimit, self.unreported)


class FamilyCode(Object):
    def __init__(self):
        self.accountID = ""
        self.familyCodeStr = ""

    def __str__(self):
        return "AccountId: %s, FamilyCodeStr: %s" % (self.accountID, self.familyCodeStr)


class PriceIncrement(Object):
    def __init__(self):
        self.lowEdge = 0.
        self.increment = 0.

    def __str__(self):
        return "LowEdge: %s, Increment: %s" % (ibapi.utils.floatMaxString(self.lowEdge), ibapi.utils.floatMaxString(self.increment))


class HistoricalTick(Object):
    def __init__(self):
        self.time = 0
        self.price = 0.
        self.size = UNSET_DECIMAL

    def __str__(self):
        return "Time: %s, Price: %s, Size: %s" % (ibapi.utils.intMaxString(self.time), ibapi.utils.floatMaxString(self.price), ibapi.utils.decimalMaxString(self.size))


class HistoricalTickBidAsk(Object):
    def __init__(self):
        self.time = 0
        self.tickAttribBidAsk = TickAttribBidAsk()
        self.priceBid = 0.
        self.priceAsk = 0.
        self.sizeBid = UNSET_DECIMAL
        self.sizeAsk = UNSET_DECIMAL

    def __str__(self):
        return "Time: %s, TickAttriBidAsk: %s, PriceBid: %s, PriceAsk: %s, SizeBid: %s, SizeAsk: %s" % (ibapi.utils.intMaxString(self.time), self.tickAttribBidAsk, 
            ibapi.utils.floatMaxString(self.priceBid), ibapi.utils.floatMaxString(self.priceAsk), 
            ibapi.utils.decimalMaxString(self.sizeBid), ibapi.utils.decimalMaxString(self.sizeAsk))


class HistoricalTickLast(Object):
    def __init__(self):
        self.time = 0
        self.tickAttribLast = TickAttribLast()
        self.price = 0.
        self.size = UNSET_DECIMAL
        self.exchange = ""
        self.specialConditions = ""

    def __str__(self):
        return "Time: %s, TickAttribLast: %s, Price: %s, Size: %s, Exchange: %s, SpecialConditions: %s" % (ibapi.utils.intMaxString(self.time), self.tickAttribLast, 
            ibapi.utils.floatMaxString(self.price), ibapi.utils.decimalMaxString(self.size), self.exchange, self.specialConditions)

class HistoricalSession(Object):
    def __init__(self):
        self.startDateTime = ""
        self.endDateTime = ""
        self.refDate = ""

    def __str__(self):
        return "Start: %s, End: %s, Ref Date: %s" % (self.startDateTime, self.endDateTime, self.refDate)

class WshEventData(Object):
    def __init__(self):
        self.conId = UNSET_INTEGER
        self.filter = ""
        self.fillWatchlist = False
        self.fillPortfolio = False
        self.fillCompetitors = False
        self.startDate = ""
        self.endDate = ""
        self.totalLimit = UNSET_INTEGER

    def __str__(self):
        return "WshEventData. ConId: %s, Filter: %s, Fill Watchlist: %d, Fill Portfolio: %d, Fill Competitors: %d" % (ibapi.utils.intMaxString(self.conId), 
            self.filter, self.fillWatchlist, self.fillPortfolio, self.fillCompetitors)




"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""

"""
The known server versions.
"""

#MIN_SERVER_VER_REAL_TIME_BARS       = 34
#MIN_SERVER_VER_SCALE_ORDERS         = 35
#MIN_SERVER_VER_SNAPSHOT_MKT_DATA    = 35
#MIN_SERVER_VER_SSHORT_COMBO_LEGS    = 35
#MIN_SERVER_VER_WHAT_IF_ORDERS       = 36
#MIN_SERVER_VER_CONTRACT_CONID       = 37
MIN_SERVER_VER_PTA_ORDERS             = 39
MIN_SERVER_VER_FUNDAMENTAL_DATA       = 40
MIN_SERVER_VER_DELTA_NEUTRAL          = 40
MIN_SERVER_VER_CONTRACT_DATA_CHAIN    = 40
MIN_SERVER_VER_SCALE_ORDERS2          = 40
MIN_SERVER_VER_ALGO_ORDERS            = 41
MIN_SERVER_VER_EXECUTION_DATA_CHAIN   = 42
MIN_SERVER_VER_NOT_HELD               = 44
MIN_SERVER_VER_SEC_ID_TYPE            = 45
MIN_SERVER_VER_PLACE_ORDER_CONID      = 46
MIN_SERVER_VER_REQ_MKT_DATA_CONID     = 47
MIN_SERVER_VER_REQ_CALC_IMPLIED_VOLAT = 49
MIN_SERVER_VER_REQ_CALC_OPTION_PRICE  = 50
MIN_SERVER_VER_SSHORTX_OLD            = 51
MIN_SERVER_VER_SSHORTX                = 52
MIN_SERVER_VER_REQ_GLOBAL_CANCEL      = 53
MIN_SERVER_VER_HEDGE_ORDERS           = 54
MIN_SERVER_VER_REQ_MARKET_DATA_TYPE   = 55
MIN_SERVER_VER_OPT_OUT_SMART_ROUTING  = 56
MIN_SERVER_VER_SMART_COMBO_ROUTING_PARAMS = 57
MIN_SERVER_VER_DELTA_NEUTRAL_CONID    = 58
MIN_SERVER_VER_SCALE_ORDERS3          = 60
MIN_SERVER_VER_ORDER_COMBO_LEGS_PRICE = 61
MIN_SERVER_VER_TRAILING_PERCENT       = 62
MIN_SERVER_VER_DELTA_NEUTRAL_OPEN_CLOSE = 66
MIN_SERVER_VER_POSITIONS              = 67
MIN_SERVER_VER_ACCOUNT_SUMMARY        = 67
MIN_SERVER_VER_TRADING_CLASS          = 68
MIN_SERVER_VER_SCALE_TABLE            = 69
MIN_SERVER_VER_LINKING                = 70
MIN_SERVER_VER_ALGO_ID                = 71
MIN_SERVER_VER_OPTIONAL_CAPABILITIES  = 72
MIN_SERVER_VER_ORDER_SOLICITED        = 73
MIN_SERVER_VER_LINKING_AUTH           = 74
MIN_SERVER_VER_PRIMARYEXCH            = 75
MIN_SERVER_VER_RANDOMIZE_SIZE_AND_PRICE = 76
MIN_SERVER_VER_FRACTIONAL_POSITIONS   = 101
MIN_SERVER_VER_PEGGED_TO_BENCHMARK    = 102
MIN_SERVER_VER_MODELS_SUPPORT         = 103
MIN_SERVER_VER_SEC_DEF_OPT_PARAMS_REQ = 104
MIN_SERVER_VER_EXT_OPERATOR           = 105
MIN_SERVER_VER_SOFT_DOLLAR_TIER       = 106
MIN_SERVER_VER_REQ_FAMILY_CODES       = 107
MIN_SERVER_VER_REQ_MATCHING_SYMBOLS   = 108
MIN_SERVER_VER_PAST_LIMIT             = 109
MIN_SERVER_VER_MD_SIZE_MULTIPLIER     = 110
MIN_SERVER_VER_CASH_QTY               = 111
MIN_SERVER_VER_REQ_MKT_DEPTH_EXCHANGES = 112
MIN_SERVER_VER_TICK_NEWS              = 113
MIN_SERVER_VER_REQ_SMART_COMPONENTS   = 114
MIN_SERVER_VER_REQ_NEWS_PROVIDERS     = 115
MIN_SERVER_VER_REQ_NEWS_ARTICLE       = 116
MIN_SERVER_VER_REQ_HISTORICAL_NEWS    = 117
MIN_SERVER_VER_REQ_HEAD_TIMESTAMP     = 118
MIN_SERVER_VER_REQ_HISTOGRAM          = 119
MIN_SERVER_VER_SERVICE_DATA_TYPE      = 120
MIN_SERVER_VER_AGG_GROUP              = 121
MIN_SERVER_VER_UNDERLYING_INFO        = 122
MIN_SERVER_VER_CANCEL_HEADTIMESTAMP   = 123
MIN_SERVER_VER_SYNT_REALTIME_BARS     = 124
MIN_SERVER_VER_CFD_REROUTE            = 125
MIN_SERVER_VER_MARKET_RULES           = 126
MIN_SERVER_VER_PNL                    = 127
MIN_SERVER_VER_NEWS_QUERY_ORIGINS     = 128
MIN_SERVER_VER_UNREALIZED_PNL         = 129
MIN_SERVER_VER_HISTORICAL_TICKS       = 130
MIN_SERVER_VER_MARKET_CAP_PRICE       = 131
MIN_SERVER_VER_PRE_OPEN_BID_ASK       = 132
MIN_SERVER_VER_REAL_EXPIRATION_DATE   = 134
MIN_SERVER_VER_REALIZED_PNL           = 135
MIN_SERVER_VER_LAST_LIQUIDITY         = 136
MIN_SERVER_VER_TICK_BY_TICK           = 137
MIN_SERVER_VER_DECISION_MAKER         = 138
MIN_SERVER_VER_MIFID_EXECUTION        = 139
MIN_SERVER_VER_TICK_BY_TICK_IGNORE_SIZE = 140
MIN_SERVER_VER_AUTO_PRICE_FOR_HEDGE     = 141
MIN_SERVER_VER_WHAT_IF_EXT_FIELDS       = 142
MIN_SERVER_VER_SCANNER_GENERIC_OPTS     = 143
MIN_SERVER_VER_API_BIND_ORDER           = 144
MIN_SERVER_VER_ORDER_CONTAINER          = 145
MIN_SERVER_VER_SMART_DEPTH              = 146
MIN_SERVER_VER_REMOVE_NULL_ALL_CASTING  = 147
MIN_SERVER_VER_D_PEG_ORDERS             = 148
MIN_SERVER_VER_MKT_DEPTH_PRIM_EXCHANGE  = 149
MIN_SERVER_VER_COMPLETED_ORDERS         = 150
MIN_SERVER_VER_PRICE_MGMT_ALGO          = 151
MIN_SERVER_VER_STOCK_TYPE               = 152
MIN_SERVER_VER_ENCODE_MSG_ASCII7        = 153
MIN_SERVER_VER_SEND_ALL_FAMILY_CODES    = 154
MIN_SERVER_VER_NO_DEFAULT_OPEN_CLOSE    = 155
MIN_SERVER_VER_PRICE_BASED_VOLATILITY   = 156
MIN_SERVER_VER_REPLACE_FA_END           = 157
MIN_SERVER_VER_DURATION                 = 158
MIN_SERVER_VER_MARKET_DATA_IN_SHARES    = 159
MIN_SERVER_VER_POST_TO_ATS              = 160
MIN_SERVER_VER_WSHE_CALENDAR            = 161
MIN_SERVER_VER_AUTO_CANCEL_PARENT       = 162
MIN_SERVER_VER_FRACTIONAL_SIZE_SUPPORT  = 163
MIN_SERVER_VER_SIZE_RULES               = 164
MIN_SERVER_VER_HISTORICAL_SCHEDULE      = 165
MIN_SERVER_VER_ADVANCED_ORDER_REJECT    = 166
MIN_SERVER_VER_USER_INFO                = 167
MIN_SERVER_VER_CRYPTO_AGGREGATED_TRADES = 168
MIN_SERVER_VER_MANUAL_ORDER_TIME        = 169
MIN_SERVER_VER_PEGBEST_PEGMID_OFFSETS   = 170
MIN_SERVER_VER_WSH_EVENT_DATA_FILTERS   = 171
MIN_SERVER_VER_IPO_PRICES               = 172
MIN_SERVER_VER_WSH_EVENT_DATA_FILTERS_DATE = 173
MIN_SERVER_VER_INSTRUMENT_TIMEZONE         = 174
MIN_SERVER_VER_HMDS_MARKET_DATA_IN_SHARES  = 175
MIN_SERVER_VER_BOND_ISSUERID               = 176


# 100+ messaging */
# 100 = enhanced handshake, msg length prefixes

MIN_CLIENT_VER = 100
MAX_CLIENT_VER = MIN_SERVER_VER_BOND_ISSUERID




"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""

from .common import UNSET_DOUBLE


class OrderState:
    def __init__(self):
        self.status= ""

        self.initMarginBefore= ""
        self.maintMarginBefore= ""
        self.equityWithLoanBefore= ""
        self.initMarginChange= ""
        self.maintMarginChange= ""
        self.equityWithLoanChange= ""
        self.initMarginAfter= ""
        self.maintMarginAfter= ""
        self.equityWithLoanAfter= ""

        self.commission = UNSET_DOUBLE      # type: float
        self.minCommission = UNSET_DOUBLE   # type: float
        self.maxCommission = UNSET_DOUBLE   # type: float
        self.commissionCurrency = ""
        self.warningText = ""
        self.completedTime = ""
        self.completedStatus = ""



from datetime import datetime, timedelta
import pandas as pd
import numpy as np
from munch import Munch
from util_py.security import handify
import math
import random

class TickManager():
    def __init__(self, main_trade) -> None:
        self.main_trade = main_trade

        self.symbol_req_map = {}
        self.req_symbol_map = {}

        self.symbol_datetime_map = {} # 最新的时间
        self.symbol_price_map = {} # 最新的成交价
        self.symbol_last_volume_map = {} # 最近的成交量

        self.symbol_ticks_map = {} # 当日历史ticks

        self.symbol_bid_map = {} # 当买一盘口
        self.symbol_ask_map = {} # 当卖一盘口
        self.symbol_order_book_map = {} # 集合买一和卖一

        self.symbol_pre_close_map = {} # 昨收
        self.symbol_volume_map = {} # 当日累计成交量
        self.symbol_quote_map = {} # 当天日行情

        self.symbol_rt_data_map = {} # 用于分时数据

    def bind(self, symbol, req_id):
        self.symbol_req_map[symbol] = req_id
        self.req_symbol_map[req_id] = symbol

    def update_time(self, req_id, timestamp):
        symbol = self.req_symbol_map.get(req_id)
        if not symbol:
            return
        
        self.symbol_datetime_map[symbol] = datetime.fromtimestamp(int(timestamp))
        self.update_tick(req_id)

    def update_pre_close(self, req_id, price):
        symbol = self.req_symbol_map.get(req_id)
        if not symbol:
            return
        
        self.symbol_pre_close_map[symbol] = price
        self.update_tick(req_id)

    def update_volume(self, req_id, volume):
        symbol = self.req_symbol_map.get(req_id)
        if not symbol:
            return
        
        self.symbol_volume_map[symbol] = float(volume)
        self.update_tick(req_id)

    def update_last_price(self, req_id, price):
        symbol = self.req_symbol_map.get(req_id)
        if not symbol:
            return
        
        self.symbol_price_map[symbol] = price
        self.update_tick(req_id)
        
    def update_last_volume(self, req_id, volume):
        symbol = self.req_symbol_map.get(req_id)
        if not symbol:
            return
        
        self.symbol_last_volume_map[symbol] = float(volume)
        self.update_tick(req_id)

    def update_bid(self, req_id, price):
        symbol = self.req_symbol_map.get(req_id)
        if not symbol:
            return
        
        self.symbol_bid_map[symbol] = price
        self.update_order_book(req_id)

    def update_ask(self, req_id, price):
        symbol = self.req_symbol_map.get(req_id)
        if not symbol:
            return
        
        self.symbol_ask_map[symbol] = price
        self.update_order_book(req_id)

    def update_order_book(self, req_id):
        symbol = self.req_symbol_map.get(req_id)
        if not symbol:
            return
        
        bid = self.symbol_bid_map.get(symbol)
        ask = self.symbol_ask_map.get(symbol)

        if bid is None or ask is None:
            return
        
        self.symbol_order_book_map[symbol] = Munch.fromDict({
            'symbol': symbol,
            'asks': [{'price': ask}],
            'bids': [{'price': bid }]
        })

        # print(f'self.symbol_order_book_map[{symbol}]', self.symbol_order_book_map[symbol])
        self.main_trade.on_order_book(self.symbol_order_book_map[symbol])
    
    def update_tick(self, req_id):
        symbol = self.req_symbol_map.get(req_id)
        if not symbol:
            return
        
        detail = self.main_trade.contract_manager.detail(symbol)
        if not detail:
            return
        
        datetime = self.symbol_datetime_map.get(symbol)
        last_price = self.symbol_price_map.get(symbol)
        last_volume = self.symbol_last_volume_map.get(symbol)
        pre_close = self.symbol_pre_close_map.get(symbol)
        volume = self.symbol_volume_map.get(symbol)
        
        if datetime is None or last_price is None or last_volume is None or pre_close is None or volume is None:
            return
        
        if self.symbol_ticks_map.get(symbol) is None:
            self.symbol_ticks_map[symbol] = pd.DataFrame()

        # 如果最后一个时间相同，则更新值，如果时间戳不同，则增加值
        new_tick = pd.DataFrame({
            'symbol': [symbol],
            'datetime': [datetime],
            'volume': [volume],
            'last_price': [last_price],
            'last_volume': [last_volume],
            'pre_close': [pre_close],
            'per_volume_random': 0
        })
        self.symbol_ticks_map[symbol] = pd.concat([self.symbol_ticks_map[symbol], new_tick], ignore_index=True)[:30]

        # 因为要更新新的tick，所以要标记过时
        del(self.symbol_datetime_map[symbol])
        del(self.symbol_price_map[symbol])
        del(self.symbol_last_volume_map[symbol])
        
        # 计算只能算法拆单量
        # lot_size = 1 if detail.stock_type == SecurityType.OPT else detail.info.lot_size # todo期权要检查
        lot_size = float(detail.lot_size)
        per_volume = handify(self.symbol_ticks_map[symbol]['last_volume'].mean(), lot_size)
        bias = 0.15
        hands = per_volume/lot_size
        part_hands = math.floor(hands*bias)
        random_hands = random.randint(-part_hands, part_hands)
        random_volume = random_hands * lot_size
        market_volume_random = per_volume + random_volume
        per_volume_random = handify(market_volume_random, lot_size)
        self.symbol_ticks_map[symbol].at[len(self.symbol_ticks_map[symbol])-1, 'per_volume_random'] = per_volume_random

        new_tick = self.symbol_ticks_map[symbol].iloc[-1]  
        self.main_trade.on_tick(new_tick)

        self.update_realtime(new_tick)


    def ticks(self, symbol):
        ticks = self.symbol_ticks_map.get(symbol)
        if ticks is None:
            return pd.DataFrame(columns=['symbol', 'datetime', 'volume', 'last_price', 'last_volume', 'pre_close'])
        
        return ticks
    
    def order_book(self, symbol):
        return self.symbol_order_book_map.get(symbol)
    
    def update_realtime(self, tick):
        symbol = tick.symbol
        volume = tick.volume
        # 计算分时数据
        next_time = (tick.datetime+timedelta(minutes=1)).replace(second=0, microsecond=0)
        time = next_time.strftime('%H:%M') # quote的是当前时间，而rt_data的时间应该是当前这一分钟的结束，也就是下一分钟整的开始 
        rt_data = self.symbol_rt_data_map.get(symbol)
        if not rt_data:
            self.symbol_rt_data_map[symbol] = rt_data = Munch.fromDict({'_volume': [('init_time', 0), (time, volume)]}) # 0元素是上一分钟，1元素是这一分钟
        else:
            # 如果当前这一分钟已经变化，则要更新这一分钟当前值和前一分钟的最后值；否则只更新当前一分钟值
            if not rt_data._volume[1][0] == time:
                rt_data._volume[0] = rt_data._volume[1]
            rt_data._volume[1] = (time, volume)
        rt_data.volume = float(rt_data._volume[1][1] - rt_data._volume[0][1])
        rt_data.datetime = next_time

    def rt_data(self, symbol):
        return self.symbol_rt_data_map.get(symbol)



"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""


"""
	SAME_POS    = open/close leg value is same as combo
	OPEN_POS    = open
	CLOSE_POS   = close
	UNKNOWN_POS = unknown
"""

from .object_implem import Object
from .common import UNSET_DECIMAL
from .utils import intMaxString
from .utils import floatMaxString
from .utils import decimalMaxString


(SAME_POS, OPEN_POS, CLOSE_POS, UNKNOWN_POS) = range(4)


class ComboLeg(Object):
    def __init__(self):
        self.conId = 0  # type: int
        self.ratio = 0  # type: int
        self.action = ""      # BUY/SELL/SSHORT
        self.exchange = ""
        self.openClose = 0   # type: int; LegOpenClose enum values
        # for stock legs when doing short sale
        self.shortSaleSlot = 0
        self.designatedLocation = ""
        self.exemptCode = -1


    def __str__(self):
        return ",".join((
            intMaxString(self.conId),
            intMaxString(self.ratio),
            str(self.action),
            str(self.exchange),
            intMaxString(self.openClose),
            intMaxString(self.shortSaleSlot),
            str(self.designatedLocation),
            intMaxString(self.exemptCode)))


class DeltaNeutralContract(Object):
    def __init__(self):
        self.conId = 0   # type: int
        self.delta = 0.  # type: float
        self.price = 0.  # type: float

    def __str__(self):
        return ",".join((
            str(self.conId),
            floatMaxString(self.delta),
            floatMaxString(self.price)))


class Contract(Object):
    def __init__(self):
        self.conId = 0
        self.symbol = ""
        self.secType = ""
        self.lastTradeDateOrContractMonth = ""
        self.strike = 0.  # float !!
        self.right = ""
        self.multiplier = ""
        self.exchange = ""
        self.primaryExchange = "" # pick an actual (ie non-aggregate) exchange that the contract trades on.  DO NOT SET TO SMART.
        self.currency = ""
        self.localSymbol = ""
        self.tradingClass = ""
        self.includeExpired = False
        self.secIdType = ""	  # CUSIP;SEDOL;ISIN;RIC
        self.secId = ""
        self.description = ""
        self.issuerId = ""

        #combos
        self.comboLegsDescrip = ""  # type: str; received in open order 14 and up for all combos
        self.comboLegs = None     # type: list<ComboLeg>
        self.deltaNeutralContract = None


    def __str__(self):
        s = ",".join((
            str(self.conId),
            str(self.symbol),
            str(self.secType),
            str(self.lastTradeDateOrContractMonth),
            floatMaxString(self.strike),
            str(self.right),
            str(self.multiplier),
            str(self.exchange),
            str(self.primaryExchange),
            str(self.currency),
            str(self.localSymbol),
            str(self.tradingClass),
            str(self.includeExpired),
            str(self.secIdType),
            str(self.secId),
            str(self.description),
            str(self.issuerId)))
        s += "combo:" + self.comboLegsDescrip

        if self.comboLegs:
            for leg in self.comboLegs:
                s += ";" + str(leg)

        if self.deltaNeutralContract:
            s += ";" + str(self.deltaNeutralContract)

        return s


class ContractDetails(Object):
    def __init__(self):
        self.contract = Contract()
        self.marketName = ""
        self.minTick = 0.
        self.orderTypes = ""
        self.validExchanges = ""
        self.priceMagnifier = 0
        self.underConId = 0
        self.longName = ""
        self.contractMonth = ""
        self.industry = ""
        self.category = ""
        self.subcategory = ""
        self.timeZoneId = ""
        self.tradingHours = ""
        self.liquidHours = ""
        self.evRule = ""
        self.evMultiplier = 0
        self.aggGroup = 0
        self.underSymbol = ""
        self.underSecType = ""
        self.marketRuleIds = ""
        self.secIdList = None
        self.realExpirationDate = ""
        self.lastTradeTime = ""
        self.stockType = ""
        self.minSize = UNSET_DECIMAL
        self.sizeIncrement = UNSET_DECIMAL
        self.suggestedSizeIncrement = UNSET_DECIMAL
        # BOND values
        self.cusip = ""
        self.ratings = ""
        self.descAppend = ""
        self.bondType = ""
        self.couponType = ""
        self.callable = False
        self.putable = False
        self.coupon = 0
        self.convertible = False
        self.maturity = ""
        self.issueDate = ""
        self.nextOptionDate = ""
        self.nextOptionType = ""
        self.nextOptionPartial = False
        self.notes = ""

    def __str__(self):
        s = ",".join((
            str(self.contract),
            str(self.marketName),
            floatMaxString(self.minTick),
            str(self.orderTypes),
            str(self.validExchanges),
            intMaxString(self.priceMagnifier),
            intMaxString(self.underConId),
            str(self.longName),
            str(self.contractMonth),
            str(self.industry),
            str(self.category),
            str(self.subcategory),
            str(self.timeZoneId),
            str(self.tradingHours),
            str(self.liquidHours),
            str(self.evRule),
            intMaxString(self.evMultiplier),
            str(self.underSymbol),
            str(self.underSecType),
            str(self.marketRuleIds),
            intMaxString(self.aggGroup),
            str(self.secIdList),
            str(self.realExpirationDate),
            str(self.stockType),
            str(self.cusip),
            str(self.ratings),
            str(self.descAppend),
            str(self.bondType),
            str(self.couponType),
            str(self.callable),
            str(self.putable),
            str(self.coupon),
            str(self.convertible),
            str(self.maturity),
            str(self.issueDate),
            str(self.nextOptionDate),
            str(self.nextOptionType),
            str(self.nextOptionPartial),
            str(self.notes),
            decimalMaxString(self.minSize),
            decimalMaxString(self.sizeIncrement),
            decimalMaxString(self.suggestedSizeIncrement)))
            
        return s


class ContractDescription(Object):
    def __init__(self):
        self.contract = Contract()
        self.derivativeSecTypes = None   # type: list of strings





from time import sleep
import asyncio

class RequestManager():
    def __init__(self) -> None:
        self.results = {}

        self.req_id = 9999
    
    def set(self, tag, req_id, result):
        if not self.results.get(tag):
            self.results[tag] = {}
        self.results[tag][req_id] = result

    def get(self, tag, req_id):
        if not self.results.get(tag):
            return None

        result = self.results.get(tag).get(req_id)
        if not result:
            return None
        
        del(self.results.get(tag)[req_id])
        return result

    # async def loop_get(self, tag, req_id):
    #     while True:
    #         # breakpoint()
    #         print('self.results.get(tag)==>', self.results.get(tag))
    #         if self.results.get(tag):
    #             result = self.results.get(tag).get(req_id)
    #             if result:
    #                 del(self.results.get(tag)[req_id])
    #                 return result
    #         asyncio.sleep(1)
    
    def next_id(self):
        self.req_id += 1
        return self.req_id
    
class Tags:
    CONTRACT_DETAILS = 'CONTRACT_DETAILS'


"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""

from .ibapi.contract import * # @UnusedWildImport


class ContractSamples:

    """ Usually, the easiest way to define a Stock/CASH contract is through 
    these four attributes.  """

    @staticmethod
    def COIN():
        #! [cashcontract]
        contract = Contract()
        contract.symbol = "F"
        contract.secType = "OPT"
        contract.currency = "USD"
        contract.exchange = "SMART"
        # contract.PrimaryExch = "NASDAQ"
        #! [cashcontract]
        return contract
    
    # HKEX
    @staticmethod
    def TX():
        #! [cashcontract]
        contract = Contract()
        contract.symbol = "00700"
        contract.secType = "STK"
        contract.currency = "HKD"
        contract.exchange = "HKEX"
        contract.PrimaryExch = "HKEX"
        #! [cashcontract]
        return contract

    @staticmethod
    def HK():
        #! [cashcontract]
        contract = Contract()
        contract.symbol = "700"
        contract.secType = "STK"
        contract.currency = "HKD"
        contract.exchange = "SEHK"
        #! [cashcontract]
        return contract

    @staticmethod
    def EurGbpFx():
        #! [cashcontract]
        contract = Contract()
        contract.symbol = "EUR"
        contract.secType = "CASH"
        contract.currency = "GBP"
        contract.exchange = "IDEALPRO"
        #! [cashcontract]
        return contract


    @staticmethod
    def Index():
        #! [indcontract]
        contract = Contract()
        contract.symbol = "DAX"
        contract.secType = "IND"
        contract.currency = "EUR"
        contract.exchange = "EUREX"
        #! [indcontract]
        return contract


    @staticmethod
    def CFD():
        #! [cfdcontract]
        contract = Contract()
        contract.symbol = "IBDE30"
        contract.secType = "CFD"
        contract.currency = "EUR"
        contract.exchange = "SMART"
        #! [cfdcontract]
        return contract


    @staticmethod
    def EuropeanStock():
        contract = Contract()
        contract.symbol = "BMW"
        contract.secType = "STK"
        contract.currency = "EUR"
        contract.exchange = "SMART"
        contract.primaryExchange = "IBIS"
        return contract

    @staticmethod
    def EuropeanStock2():
        contract = Contract()
        contract.symbol = "NOKIA"
        contract.secType = "STK"
        contract.currency = "EUR"
        contract.exchange = "SMART"
        contract.primaryExchange = "HEX"
        return contract

    @staticmethod
    def OptionAtIse():
        contract = Contract()
        contract.symbol = "COF"
        contract.secType = "OPT"
        contract.currency = "USD"
        contract.exchange = "ISE"
        contract.lastTradeDateOrContractMonth = "20190315"
        contract.right = "P"
        contract.strike = 105
        contract.multiplier = "100"
        return contract


    @staticmethod
    def BondWithCusip():
            #! [bondwithcusip]
            contract = Contract()
            # enter CUSIP as symbol
            contract.symbol= "912828C57"
            contract.secType = "BOND"
            contract.exchange = "SMART"
            contract.currency = "USD"
            #! [bondwithcusip]
            return contract


    @staticmethod
    def Bond():
            #! [bond]
            contract = Contract()
            contract.conId = 15960357
            contract.exchange = "SMART"
            #! [bond]
            return contract


    @staticmethod
    def MutualFund():
            #! [fundcontract]
            contract = Contract()
            contract.symbol = "VINIX"
            contract.secType = "FUND"
            contract.exchange = "FUNDSERV"
            contract.currency = "USD"
            #! [fundcontract]
            return contract


    @staticmethod
    def Commodity():
            #! [commoditycontract]
            contract = Contract()
            contract.symbol = "XAUUSD"
            contract.secType = "CMDTY"
            contract.exchange = "SMART"
            contract.currency = "USD"
            #! [commoditycontract]
            return contract
    

    @staticmethod
    def USStock():
        #! [stkcontract]
        contract = Contract()
        contract.symbol = "SPY"
        contract.secType = "STK"
        contract.currency = "USD"
        contract.exchange = "ARCA"
        #! [stkcontract]
        return contract


    @staticmethod
    def USStockWithPrimaryExch():
        #! [stkcontractwithprimary]
        contract = Contract()
        contract.symbol = "SPY"
        contract.secType = "STK"
        contract.currency = "USD"
        contract.exchange = "SMART"
        contract.primaryExchange = "ARCA"
        #! [stkcontractwithprimary]
        return contract

            
    @staticmethod
    def USStockAtSmart():
        contract = Contract()
        contract.symbol = "IBM"
        contract.secType = "STK"
        contract.currency = "USD"
        contract.exchange = "SMART"
        return contract

    @staticmethod
    def etf():
        contract = Contract()
        contract.symbol = "QQQ"
        contract.secType = "STK"
        contract.currency = "USD"
        contract.exchange = "SMART"
        return contract

    @staticmethod
    def USOptionContract():
        #! [optcontract_us]
        contract = Contract()
        contract.symbol = "GOOG"
        contract.secType = "OPT"
        contract.exchange = "SMART"
        contract.currency = "USD"
        contract.lastTradeDateOrContractMonth = "20190315"
        contract.strike = 1180
        contract.right = "C"
        contract.multiplier = "100"
        #! [optcontract_us]
        return contract


    @staticmethod
    def OptionAtBOX():
        #! [optcontract]
        contract = Contract()
        contract.symbol = "GOOG"
        contract.secType = "OPT"
        contract.exchange = "BOX"
        contract.currency = "USD"
        contract.lastTradeDateOrContractMonth = "20190315"
        contract.strike = 1180
        contract.right = "C"
        contract.multiplier = "100"
        #! [optcontract]
        return contract


    """ Option contracts require far more information since there are many 
    contracts having the exact same attributes such as symbol, currency, 
    strike, etc. This can be overcome by adding more details such as the 
    trading class"""

    @staticmethod
    def OptionWithTradingClass():
        #! [optcontract_tradingclass]
        contract = Contract()
        contract.symbol = "SANT"
        contract.secType = "OPT"
        contract.exchange = "MEFFRV"
        contract.currency = "EUR"
        contract.lastTradeDateOrContractMonth = "20190621"
        contract.strike = 7.5
        contract.right = "C"
        contract.multiplier = "100"
        contract.tradingClass = "SANEU"
        #! [optcontract_tradingclass]
        return contract


    """ Using the contract's own symbol (localSymbol) can greatly simplify a
    contract description """

    @staticmethod
    def OptionWithLocalSymbol():
        #! [optcontract_localsymbol]
        contract = Contract()
        #Watch out for the spaces within the local symbol!
        contract.localSymbol = "P BMW  20221216 72 M"
        contract.secType = "OPT"
        contract.exchange = "EUREX"
        contract.currency = "EUR"
        #! [optcontract_localsymbol]
        return contract

    """ Dutch Warrants (IOPTs) can be defined using the local symbol or conid 
    """

    @staticmethod
    def DutchWarrant():
        #! [ioptcontract]
        contract = Contract()
        contract.localSymbol = "B881G"
        contract.secType = "IOPT"
        contract.exchange = "SBF"
        contract.currency = "EUR"
        #! [ioptcontract]
        return contract

    """ Future contracts also require an expiration date but are less
    complicated than options."""

    @staticmethod
    def SimpleFuture():
        #! [futcontract]
        contract = Contract()
        contract.symbol = "GBL"
        contract.secType = "FUT"
        contract.exchange = "EUREX"
        contract.currency = "EUR"
        contract.lastTradeDateOrContractMonth = "202303"
        #! [futcontract]
        return contract


    """Rather than giving expiration dates we can also provide the local symbol
    attributes such as symbol, currency, strike, etc. """

    @staticmethod
    def FutureWithLocalSymbol():
        #! [futcontract_local_symbol]
        contract = Contract()
        contract.secType = "FUT"
        contract.exchange = "EUREX"
        contract.currency = "EUR"
        contract.localSymbol = "FGBL MAR 23"
        #! [futcontract_local_symbol]
        return contract


    @staticmethod
    def FutureWithMultiplier():
        #! [futcontract_multiplier]
        contract = Contract()
        contract.symbol = "DAX"
        contract.secType = "FUT"
        contract.exchange = "EUREX"
        contract.currency = "EUR"
        contract.lastTradeDateOrContractMonth = "202303"
        contract.multiplier = "1"
        #! [futcontract_multiplier]
        return contract


    """ Note the space in the symbol! """

    @staticmethod
    def WrongContract():
        contract = Contract()
        contract.symbol = " IJR "
        contract.conId = 9579976
        contract.secType = "STK"
        contract.exchange = "SMART"
        contract.currency = "USD"
        return contract

    @staticmethod
    def FuturesOnOptions():
        #! [fopcontract]
        contract = Contract()
        contract.symbol = "GBL"
        contract.secType = "FOP"
        contract.exchange = "EUREX"
        contract.currency = "EUR"
        contract.lastTradeDateOrContractMonth = "20230224"
        contract.strike = 138
        contract.right = "C"
        contract.multiplier = "1000"
        #! [fopcontract]
        return contract

    @staticmethod
    def Warrants():
        #! [warcontract]
        contract = Contract()
        contract.symbol = "GOOG"
        contract.secType = "WAR"
        contract.exchange = "FWB"
        contract.currency = "EUR"
        contract.lastTradeDateOrContractMonth = "20201117"
        contract.strike = 1500.0
        contract.right = "C"
        contract.multiplier = "0.01"
        #! [warcontract]
        return contract

    """ It is also possible to define contracts based on their ISIN (IBKR STK
    sample). """

    @staticmethod
    def ByISIN():
        contract = Contract()
        contract.secIdType = "ISIN"
        contract.secId = "US45841N1072"
        contract.exchange = "SMART"
        contract.currency = "USD"
        contract.secType = "STK"
        return contract


    """ Or their conId (EUR.uSD sample).
    Note: passing a contract containing the conId can cause problems if one of 
    the other provided attributes does not match 100% with what is in IB's 
    database. This is particularly important for contracts such as Bonds which 
    may change their description from one day to another.
    If the conId is provided, it is best not to give too much information as
    in the example below. """

    @staticmethod
    def ByConId():
        contract = Contract()
        contract.secType = "CASH"
        contract.conId = 12087792
        contract.exchange = "IDEALPRO"
        return contract


    """ Ambiguous contracts are great to use with reqContractDetails. This way
    you can query the whole option chain for an underlying. Bear in mind that
    there are pacing mechanisms in place which will delay any further responses
    from the TWS to prevent abuse. """

    @staticmethod
    def COIN_OPT():
        #! [optionforquery]
        contract = Contract()
        contract.symbol = "COIN"
        contract.secType = "OPT"
        contract.exchange = "SMART"
        contract.currency = "USD"
        contract.lastTradeDateOrContractMonth = '20230915'
        contract.strike = 82
        contract.right = 'C'
        #! [optionforquery]
        return contract


    @staticmethod
    def OptionComboContract():
        #! [bagoptcontract]
        contract = Contract()
        contract.symbol = "DBK"
        contract.secType = "BAG"
        contract.currency = "EUR"
        contract.exchange = "EUREX"

        leg1 = ComboLeg()
        leg1.conId = 577164786 #DBK Jun21'24 2 CALL @EUREX
        leg1.ratio = 1
        leg1.action = "BUY"
        leg1.exchange = "EUREX"

        leg2 = ComboLeg()
        leg2.conId = 577164767 #DBK Dec15'23 2 CALL @EUREX
        leg2.ratio = 1
        leg2.action = "SELL"
        leg2.exchange = "EUREX"

        contract.comboLegs = []
        contract.comboLegs.append(leg1)
        contract.comboLegs.append(leg2)
        #! [bagoptcontract]
        return contract


    """ STK Combo contract
    Leg 1: 43645865 - IBKR's STK
    Leg 2: 9408 - McDonald's STK """

    @staticmethod
    def StockComboContract():
        #! [bagstkcontract]
        contract = Contract()
        contract.symbol = "IBKR,MCD"
        contract.secType = "BAG"
        contract.currency = "USD"
        contract.exchange = "SMART"

        leg1 = ComboLeg()
        leg1.conId = 43645865#IBKR STK
        leg1.ratio = 1
        leg1.action = "BUY"
        leg1.exchange = "SMART"

        leg2 = ComboLeg()
        leg2.conId = 9408#MCD STK
        leg2.ratio = 1
        leg2.action = "SELL"
        leg2.exchange = "SMART"

        contract.comboLegs = []
        contract.comboLegs.append(leg1)
        contract.comboLegs.append(leg2)
        #! [bagstkcontract]
        return contract


    """ CBOE Volatility Index Future combo contract """

    @staticmethod
    def FutureComboContract():
        #! [bagfutcontract]
        contract = Contract()
        contract.symbol = "VIX"
        contract.secType = "BAG"
        contract.currency = "USD"
        contract.exchange = "CFE"

        leg1 = ComboLeg()
        leg1.conId = 326501438 # VIX FUT 201903
        leg1.ratio = 1
        leg1.action = "BUY"
        leg1.exchange = "CFE"

        leg2 = ComboLeg()
        leg2.conId = 323072528 # VIX FUT 2019049
        leg2.ratio = 1
        leg2.action = "SELL"
        leg2.exchange = "CFE"

        contract.comboLegs = []
        contract.comboLegs.append(leg1)
        contract.comboLegs.append(leg2)
        #! [bagfutcontract]
        return contract

    @staticmethod
    def SmartFutureComboContract():
        #! [smartfuturespread]
        contract = Contract()
        contract.symbol = "WTI" # WTI,COIL spread. Symbol can be defined as first leg symbol ("WTI") or currency ("USD")
        contract.secType = "BAG"
        contract.currency = "USD"
        contract.exchange = "SMART"

        leg1 = ComboLeg()
        leg1.conId = 55928698 # WTI future June 2017
        leg1.ratio = 1
        leg1.action = "BUY"
        leg1.exchange = "IPE"

        leg2 = ComboLeg()
        leg2.conId = 55850663 # COIL future June 2017
        leg2.ratio = 1
        leg2.action = "SELL"
        leg2.exchange = "IPE"

        contract.comboLegs = []
        contract.comboLegs.append(leg1)
        contract.comboLegs.append(leg2)
        #! [smartfuturespread]
        return contract

    @staticmethod
    def InterCmdtyFuturesContract():
        #! [intcmdfutcontract]
        contract = Contract()
        contract.symbol = "COL.WTI" #symbol is 'local symbol' of intercommodity spread. 
        contract.secType = "BAG"
        contract.currency = "USD"
        contract.exchange = "IPE"

        leg1 = ComboLeg()
        leg1.conId = 183405603 #WTI�Dec'23�@IPE
        leg1.ratio = 1
        leg1.action = "BUY"
        leg1.exchange = "IPE"

        leg2 = ComboLeg()
        leg2.conId = 254011009 #COIL�Dec'23�@IPE
        leg2.ratio = 1
        leg2.action = "SELL"
        leg2.exchange = "IPE"

        contract.comboLegs = []
        contract.comboLegs.append(leg1)
        contract.comboLegs.append(leg2)
        #! [intcmdfutcontract]
        return contract


    @staticmethod
    def NewsFeedForQuery():
        #! [newsfeedforquery]
        contract = Contract()
        contract.secType = "NEWS"
        contract.exchange = "BRFG" #Briefing Trader
        #! [newsfeedforquery]
        return contract


    @staticmethod
    def BTbroadtapeNewsFeed():
        #! [newscontractbt]
        contract = Contract()
        contract.symbol  = "BRF:BRF_ALL"
        contract.secType = "NEWS"
        contract.exchange = "BRF"
        #! [newscontractbt]
        return contract


    @staticmethod
    def BZbroadtapeNewsFeed():
        #! [newscontractbz]
        contract = Contract()
        contract.symbol = "BZ:BZ_ALL"
        contract.secType = "NEWS"
        contract.exchange = "BZ"
        #! [newscontractbz]
        return contract


    @staticmethod
    def FLYbroadtapeNewsFeed():
        #! [newscontractfly]
        contract = Contract()
        contract.symbol  = "FLY:FLY_ALL"
        contract.secType = "NEWS"
        contract.exchange = "FLY"
        #! [newscontractfly]
        return contract


    @staticmethod
    def ContFut():
        #! [continuousfuturescontract]
        contract = Contract()
        contract.symbol = "GBL"
        contract.secType = "CONTFUT"
        contract.exchange = "EUREX"
        #! [continuousfuturescontract]
        return contract

    @staticmethod
    def ContAndExpiringFut():
        #! [contandexpiringfut]
        contract = Contract()
        contract.symbol = "GBL"
        contract.secType = "FUT+CONTFUT"
        contract.exchange = "EUREX"
        #! [contandexpiringfut]
        return contract

    @staticmethod
    def JefferiesContract():
        #! [jefferies_contract]
        contract = Contract()
        contract.symbol = "AAPL"
        contract.secType = "STK"
        contract.exchange = "JEFFALGO"
        contract.currency = "USD"
        #! [jefferies_contract]
        return contract

    @staticmethod
    def CSFBContract():
        #! [csfb_contract]
        contract = Contract()
        contract.symbol = "IBKR"
        contract.secType = "STK"
        contract.exchange = "CSFBALGO"
        contract.currency = "USD"
        #! [csfb_contract]
        return contract

    @staticmethod
    def USStockCFD():
        # ! [usstockcfd_conract]
        contract = Contract();
        contract.symbol = "IBM";
        contract.secType = "CFD";
        contract.currency = "USD";
        contract.exchange = "SMART";
        # ! [usstockcfd_conract]
        return contract;

    @staticmethod
    def EuropeanStockCFD():
        # ! [europeanstockcfd_contract]
        contract = Contract();
        contract.symbol = "BMW";
        contract.secType = "CFD";
        contract.currency = "EUR";
        contract.exchange = "SMART";
        # ! [europeanstockcfd_contract]
        return contract;

    @staticmethod
    def CashCFD():
        # ! [cashcfd_contract]
        contract = Contract();
        contract.symbol = "EUR";
        contract.secType = "CFD";
        contract.currency = "USD";
        contract.exchange = "SMART";
        # ! [cashcfd_contract]
        return contract;

    @staticmethod
    def QBAlgoContract():
        # ! [qbalgo_contract]
        contract = Contract()
        contract.symbol = "ES";
        contract.secType = "FUT";
        contract.exchange = "QBALGO";
        contract.currency = "USD";
        contract.lastTradeDateOrContractMonth = "202003";
        # ! [qbalgo_contract]
        return contract; 

    @staticmethod
    def IBKRATSContract():
        # ! [ibkrats_contract]
        contract = Contract();
        contract.symbol = "SPY";
        contract.secType = "STK";
        contract.currency = "USD";
        contract.exchange = "IBKRATS";
        # ! [ibkrats_contract]
        return contract;

    @staticmethod
    def CryptoContract():
        # ! [crypto_contract]
        contract = Contract();
        contract.symbol = "ETH";
        contract.secType = "CRYPTO";
        contract.currency = "USD";
        contract.exchange = "PAXOS";
        # ! [crypto_contract]
        return contract;

    @staticmethod
    def StockWithIPOPrice():
        # ! [stock_with_IPO_price]
        contract = Contract()
        contract.symbol = "EMCGU";
        contract.secType = "STK";
        contract.currency = "USD";
        contract.exchange = "SMART";
        # ! [stock_with_IPO_price]
        return contract

    @staticmethod
    def ByFIGI():
        # ! [ByFIGI]
        contract = Contract()
        contract.secIdType = "FIGI";
        contract.secId = "BBG000B9XRY4";
        contract.exchange = "SMART";
        # ! [ByFIGI]
        return contract
        
    @staticmethod
    def ByIssuerId():
        # ! [ByIssuerId]
        contract = Contract()
        contract.issuerId = "e1453318";
        # ! [ByIssuerId]
        return contract

def Test():
    from .ibapi.utils import ExerciseStaticMethods
    ExerciseStaticMethods(ContractSamples)


if "__main__" == __name__:
    Test()




"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""


"""
The EReader runs in a separate threads and is responsible for receiving the
incoming messages.
It will read the packets from the wire, use the low level IB messaging to
remove the size prefix and put the rest in a Queue.
"""

import logging
from threading import Thread

from . import comm


logger = logging.getLogger(__name__)


class EReader(Thread):
    def __init__(self, conn, msg_queue):
        super().__init__()
        self.conn = conn
        self.msg_queue = msg_queue

    def run(self):
        try:
            logger.debug("EReader thread started")
            buf = b""
            while self.conn.isConnected():

                data = self.conn.recvMsg()
                logger.debug("reader loop, recvd size %d", len(data))
                buf += data

                while len(buf) > 0:
                    (size, msg, buf) = comm.read_msg(buf)
                    #logger.debug("resp %s", buf.decode('ascii'))
                    logger.debug("size:%d msg.size:%d msg:|%s| buf:%s|", size,
                        len(msg), buf, "|")

                    if msg:
                        self.msg_queue.put(msg)
                    else:
                        logger.debug("more incoming packet(s) are needed ")
                        break

            logger.debug("EReader thread finished")
        except:
            logger.exception('unhandled exception in EReader thread')




"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""

from .ibapi.object_implem import Object


class FaAllocationSamples(Object):
    #! [faonegroup]
    FaOneGroup = "".join(("<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
        , "<ListOfGroups>"                                                   
                    , "<Group>"                                              
                            , "<name>Equal_Quantity</name>"                  
                            , "<ListOfAccts varName=\"list\">"               
                                    #Replace with your own accountIds
                                    , "<String>DU119915</String>"            
                                    , "<String>DU119916</String>"            
                            , "</ListOfAccts>"                               
                            , "<defaultMethod>EqualQuantity</defaultMethod>" 
                    , "</Group>"                                             
            , "</ListOfGroups>"))
    #! [faonegroup]

    #! [fatwogroups]
    FaTwoGroups = "".join(("<?xml version=\"1.0\" encoding=\"UTF-8\"?>"               
                ,"<ListOfGroups>"                                            
                    , "<Group>"                                              
                            , "<name>Equal_Quantity</name>"                  
                            , "<ListOfAccts varName=\"list\">"               
                                    #Replace with your own accountIds                            
                                    , "<String>DU119915</String>"                                
                                    , "<String>DU119916</String>"            
                            , "</ListOfAccts>"                               
                            , "<defaultMethod>EqualQuantity</defaultMethod>" 
                    , "</Group>"                                             
                    , "<Group>"                                              
                            , "<name>Pct_Change</name>"                      
                            , "<ListOfAccts varName=\"list\">"               
                #Replace with your own accountIds                            
                , "<String>DU119915</String>"                                
                                    , "<String>DU119916</String>"            
                            , "</ListOfAccts>"                               
                            , "<defaultMethod>PctChange</defaultMethod>"     
                    , "</Group>"                                             
            , "</ListOfGroups>"))
    #! [fatwogroups]


    #! [faoneprofile]
    FaOneProfile = "".join(("<?xml version=\"1.0\" encoding=\"UTF-8\"?>"              
                , "<ListOfAllocationProfiles>"                               
                    , "<AllocationProfile>"                                  
                            , "<name>Percent_60_40</name>"                   
                            , "<type>1</type>"                               
                            , "<ListOfAllocations varName=\"listOfAllocations\">"
                                    , "<Allocation>"                         
                                            #Replace with your own accountIds                        
                                            , "<acct>DU119915</acct>"                                
                                            , "<amount>60.0</amount>"        
                                    , "</Allocation>"                        
                                    , "<Allocation>"                         
                                            #Replace with your own accountIds                        
                                            , "<acct>DU119916</acct>"                                
                                            , "<amount>40.0</amount>"        
                                    , "</Allocation>"                        
                            , "</ListOfAllocations>"                         
                    , "</AllocationProfile>"                                 
            , "</ListOfAllocationProfiles>"))
    #! [faoneprofile]

    #! [fatwoprofiles]
    FaTwoProfiles = "".join(("<?xml version=\"1.0\" encoding=\"UTF-8\"?>"             
                , "<ListOfAllocationProfiles>"                               
                    , "<AllocationProfile>"                                  
                            , "<name>Percent_60_40</name>"                   
                            , "<type>1</type>"                               
                            , "<ListOfAllocations varName=\"listOfAllocations\">" 
                                    , "<Allocation>"                         
                                            #Replace with your own accountIds                        
                                            , "<acct>DU119915</acct>"                                
                                            , "<amount>60.0</amount>"        
                                    , "</Allocation>"                        
                                    , "<Allocation>"                         
                                            #Replace with your own accountIds                        
                                            , "<acct>DU119916</acct>"                                
                                            , "<amount>40.0</amount>"        
                                    , "</Allocation>"                        
                            , "</ListOfAllocations>"                         
                    , "</AllocationProfile>"                                 
                    , "<AllocationProfile>"                                  
                            , "<name>Ratios_2_1</name>"                      
                            , "<type>2</type>"                               
                            , "<ListOfAllocations varName=\"listOfAllocations\">"
                                    , "<Allocation>"                         
                                            #Replace with your own accountIds                        
                                            , "<acct>DU119915</acct>"                                
                                            , "<amount>2.0</amount>"         
                                    , "</Allocation>"                        
                                    , "<Allocation>"                         
                                            #Replace with your own accountIds                        
                                            , "<acct>DU119916</acct>"                                
                                            , "<amount>1.0</amount>"         
                                    , "</Allocation>"                        
                            , "</ListOfAllocations>"                         
                    , "</AllocationProfile>"                                 
            , "</ListOfAllocationProfiles>"))
    #! [fatwoprofiles]


def Test():
    print(FaAllocationSamples.FaOneGroup)
    print(FaAllocationSamples.FaTwoGroups)
    print(FaAllocationSamples.FaOneProfile)
    print(FaAllocationSamples.FaTwoProfiles)

if "__main__" == __name__:
    Test()





import pandas as pd
import numpy as np
from munch import Munch
from .constant import *
from .ibapi.contract import *
import re
from .base_manager import BaseManager

class ContractManager(BaseManager):
    def __init__(self, main_trade) -> None:
        super().__init__(main_trade)

        self.symbol_detail_map = {}
        self.symbol_market_map = {}

    def contract(self, symbol):
        market = self.market(symbol)
        contract = Contract()
        contract.symbol = symbol
        contract.secType = "STK"
        contract.currency = "USD" if market == Market.US else 'HKD'
        contract.exchange = market

        # 形式：'COIN230815C82000'
        result = re.match("(\w+)(\d{6})([C|P])(\d+)", symbol)
        if result: 
            contract.secType = "OPT"
            symbol, expiry, right, strike = result.groups()
            contract.symbol = symbol
            contract.lastTradeDateOrContractMonth = '20' + expiry
            contract.right = right
            contract.strike = int(strike)/1000

        return contract

    def update_detail(self, req_id, detail):
        symbol = self.get_symbol(req_id)
        if not symbol:
            return
        
        self.symbol_detail_map[symbol] = detail
        self.main_trade.on_contract_detail(detail)

    def detail(self, symbol):
        detail = self.symbol_detail_map.get(symbol)
        if not detail:
            self.main_trade.stock(symbol)
        return detail
    
    def set_market(self, symbol, market):
        self.symbol_market_map[symbol] = market
    
    def market(self, symbol):
        return self.symbol_market_map.get(symbol) or Market.US
    
    # 股票的symbol就是symbol，期权要加上三要素
    def c2s(self, contract):
        if contract.secType == SecurityType.OPT:
            return f'{contract.symbol}{contract.lastTradeDateOrContractMonth[2:]}{contract.right}{int(contract.strike*1000)}'
        return contract.symbol


from datetime import datetime
import pandas as pd
import numpy as np
from munch import Munch

REQ_ID = 9999

class BaseManager():
    # @staticmethod
    # def id():
    #     global req_id
    #     req_id += 1
    #     return req_id
    
    def __init__(self, main_trade) -> None:
        self.main_trade = main_trade

        self.symbol_req_map = {}
        self.req_symbol_map = {}

    def next_id(self, symbol):
        global REQ_ID
        REQ_ID += 1

        self.symbol_req_map[symbol] = REQ_ID
        self.req_symbol_map[REQ_ID] = symbol

        return REQ_ID
    
    def get_symbol(self, req_id):
        return self.req_symbol_map.get(req_id)

    